use {
    crate::{
        check_program_account,
        extension::interest_bearing_mint::BasisPoints,
        instruction::{encode_instruction, TokenInstruction},
        pod::OptionalNonZeroPubkey,
    },
    bytemuck::{Pod, Zeroable},
    num_enum::{IntoPrimitive, TryFromPrimitive},
    solana_program::{
        instruction::{AccountMeta, Instruction},
        program_error::ProgramError,
        pubkey::Pubkey,
    },
    std::convert::TryInto,
};

/// Interesting-bearing mint extension instructions
#[derive(Clone, Copy, Debug, PartialEq, IntoPrimitive, TryFromPrimitive)]
#[repr(u8)]
pub enum InterestBearingMintInstruction {
    /// Initialize a new mint with interest accrual.
    ///
    /// Fails if the mint has already been initialized, so must be called before
    /// `InitializeMint`.
    ///
    /// The mint must have exactly enough space allocated for the base mint (82
    /// bytes), plus 83 bytes of padding, 1 byte reserved for the account type,
    /// then space required for this extension, plus any others.
    ///
    /// Accounts expected by this instruction:
    ///
    ///   0. `[writable]` The mint to initialize.
    ///
    /// Data expected by this instruction:
    ///   `crate::extension::interest_bearing::instruction::InitializeInstructionData`
    ///
    Initialize,
    /// Update the interest rate. Only supported for mints that include the
    /// `InterestBearingConfig` extension.
    ///
    /// Accounts expected by this instruction:
    ///
    ///   * Single authority
    ///   0. `[writable]` The mint.
    ///   1. `[signer]` The mint rate authority.
    ///
    ///   * Multisignature authority
    ///   0. `[writable]` The mint.
    ///   1. `[]` The mint's multisignature rate authority.
    ///   2. ..2+M `[signer]` M signer accounts.
    ///
    /// Data expected by this instruction:
    ///   `crate::extension::interest_bearing::BasisPoints`
    ///
    UpdateRate,
}

/// Data expected by `InterestBearing::Initialize`
#[derive(Clone, Copy, Pod, Zeroable)]
#[repr(C)]
pub struct InitializeInstructionData {
    /// The public key for the account that can update the rate
    pub rate_authority: OptionalNonZeroPubkey,
    /// The initial interest rate
    pub rate: BasisPoints,
}

/// Create an `Initialize` instruction
pub fn initialize(
    token_program_id: &Pubkey,
    mint: &Pubkey,
    rate_authority: Option<Pubkey>,
    rate: i16,
) -> Result<Instruction, ProgramError> {
    check_program_account(token_program_id)?;
    let accounts = vec![AccountMeta::new(*mint, false)];
    Ok(encode_instruction(
        token_program_id,
        accounts,
        TokenInstruction::InterestBearingMintExtension,
        InterestBearingMintInstruction::Initialize,
        &InitializeInstructionData {
            rate_authority: rate_authority.try_into()?,
            rate: rate.into(),
        },
    ))
}

/// Create an `UpdateRate` instruction
pub fn update_rate(
    token_program_id: &Pubkey,
    mint: &Pubkey,
    rate_authority: &Pubkey,
    signers: &[&Pubkey],
    rate: i16,
) -> Result<Instruction, ProgramError> {
    check_program_account(token_program_id)?;
    let mut accounts = vec![
        AccountMeta::new(*mint, false),
        AccountMeta::new_readonly(*rate_authority, signers.is_empty()),
    ];
    for signer_pubkey in signers.iter() {
        accounts.push(AccountMeta::new_readonly(**signer_pubkey, true));
    }
    Ok(encode_instruction(
        token_program_id,
        accounts,
        TokenInstruction::InterestBearingMintExtension,
        InterestBearingMintInstruction::UpdateRate,
        &BasisPoints::from(rate),
    ))
}
